/*
* PARALLEL 2D HEAT EQUATION SOLVER
*
* Copyright � 2008 Tony Saad <remineur@gmail.com>
* (Written by Tony Saad <remineur@gmail.com> for UTSI and Dr. Christian Parigger's class)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
*/

#include "heat.h"

int main(int argc, char* argv[]) {

	/***********************************************************/
	/* VARIABLES
	/***********************************************************/
	bool continueRun = 1;
	int iDummy;
	int nNodesX,
		nNodesY;
	int numberOfAdjacentPartitions;
	int ix = 0,
		jy =0, 
		iFirst,
		iLast;		
	double endSimulationTime = 0.0, 
		CFL = 0.25, 
		deltaT,
		deltaS;
	double startWallTime = 0.0,
		endWallTime = 0.0, 
		totalWallTime = 0.0, 
		maximumWallTime = 0.0;			
	ProblemInfo probInfo;

	/* initialize MPI */
	if (parallel) MPISetup(argc,argv);	
	
	/* process user input */
	if (MASTER) continueRun = processUserInput(probInfo);
	
	if (parallel) {
		
		MPI_Bcast(&continueRun,1,MPI_LOGICAL,0,WORLD);
		if (!continueRun) {
			MPI_Finalize();
			return 0;
		}
		
		MPI_Bcast(&probInfo,1,MPI_ProblemInfo,0,WORLD);
		/* start recording the wall clock time */
		startWallTime = MPI_Wtime();
	}
	
	nNodesX = probInfo.numberOfNodesX;
	nNodesY = probInfo.numberOfNodesY;	
	deltaS =(double) 1.0/probInfo.globalNumberOfNodes;		
	CFL = probInfo.CFL;
	deltaT = CFL*deltaS*deltaS/probInfo.thermalDiffusivity;

	// all nodes have 2 adjacent partitions except the Master and the Last_node - they only have 1 adjacent partition
	numberOfAdjacentPartitions = 2;
	if (MASTER || LAST_NODE) numberOfAdjacentPartitions = 1;
	
	/* memory allocation */		
	double** uOld = make2DDoubleArray(nNodesX,nNodesY);	
	initializeArray(uOld,nNodesX,nNodesY);
	
	double** uNew = make2DDoubleArray(nNodesX,nNodesY);
	setupBoundaryConditions(uNew,nNodesX,nNodesY);
	
	// allocate memory for the values to be sent	
	double* valuesToSend = makeDoubleArray(nNodesY);
	
	/* allocate memory for the arrays to be received 
	   each partition receives two arrays except the Master and the last node. */
	double** valuesToReceive = make2DDoubleArray(numberOfAdjacentPartitions,nNodesY);	
	
	/* Allocate the requests and statuses for asynchronous communication */
	MPI_Request* requests =(MPI_Request*)malloc(2*numberOfAdjacentPartitions*sizeof(MPI_Request));
	MPI_Status* statuses =(MPI_Status*)malloc(2*numberOfAdjacentPartitions*sizeof(MPI_Status));

	/***********************************************************/
	/* INITIALIZE THE NUMBER OF PARTITIONS AND THEIR INDICES
	/***********************************************************/
	// allocate an array that holds the indices of the ajdacent partitions		
	int* adjacentPartitionsIndices = makeIntArray(numberOfAdjacentPartitions);	
	int* indexOfValuesToSend = makeIntArray(numberOfAdjacentPartitions);
	if (MASTER) {
		adjacentPartitionsIndices[0] = 1;
		indexOfValuesToSend[0] = nNodesX - 2;
	} else if (LAST_NODE) {
		adjacentPartitionsIndices[0] = numberOfProcessors - 2;
		indexOfValuesToSend[0] = 1;
	} else {
		adjacentPartitionsIndices[0] = LEFT_NODE;
		adjacentPartitionsIndices[1] = RIGHT_NODE;
		indexOfValuesToSend[0] = nNodesX - 2;
		indexOfValuesToSend[1] = 1;
	}

	/***********************************************************/
	/* COMPUTATION
	/***********************************************************/
	iFirst = 1;
	iLast = probInfo.globalNumberOfNodes/numberOfProcessors;
	if (MASTER) iFirst++;
	if (LAST_NODE) iLast--;
	int iter = 0;
	double currentTime = 0;
	while(iter++*deltaT <= probInfo.endSimulationTime) {
		int reqCounter = 0;		

		if (MASTER & iter % 100 == 0) {							
			printf("Iteration: %i - Current Time = %lf s\n", iter-1,(iter-1)*deltaT);
		}

		for (ix = iFirst; ix <= iLast; ix++) {
			for (jy = 1; jy < nNodesY ;jy++){
				uNew[ix][jy] = uOld[ix][jy] + CFL*(uOld[ix+1][jy] + uOld[ix-1][jy] - 4*uOld[ix][jy] + uOld[ix][jy-1] + uOld[ix][jy+1]);
			}
		}

		for (int iAdjPart = 0; iAdjPart < numberOfAdjacentPartitions; iAdjPart++) {
			int indexLocationOfValues = indexOfValuesToSend[iAdjPart];
			int adjacentPartition = adjacentPartitionsIndices[iAdjPart];
			for (int i = 0; i<nNodesY; i++) {
				valuesToSend[i] = uNew[indexLocationOfValues][i];			
			}
			MPI_Isend(valuesToSend,nNodesY,MPI_DOUBLE,adjacentPartition,processorID,WORLD,&requests[reqCounter++]);
		}

		for (int iAdjPart = 0; iAdjPart < numberOfAdjacentPartitions; iAdjPart++) {			
			int adjacentPartition = adjacentPartitionsIndices[iAdjPart];
			MPI_Irecv(valuesToReceive[iAdjPart],nNodesY,MPI_DOUBLE,adjacentPartition,adjacentPartition,WORLD,&requests[reqCounter++]);
		}

		MPI_Waitall(2*numberOfAdjacentPartitions,requests,statuses);

		/* fill uNew with the values that were received */
		for (int i = 0; i < numberOfAdjacentPartitions; i++) {			
			int adjacentPartition = adjacentPartitionsIndices[i];
			int indexOfValues = indexOfValuesToSend[i];
			for (jy = 1; jy < nNodesY; jy++) {
				uNew[indexOfValues][jy] = valuesToReceive[i][jy];
			}
		}

		/* set uOld = uNew */
		for (ix = iFirst; ix <= iLast; ix++) {
			for (jy = 1; jy < nNodesY ;jy++){
				uOld[ix][jy] = uNew[ix][jy];
			}
		}
	}
	free(requests);
	free(statuses);
		
	if (parallel) endWallTime = MPI_Wtime();
	
	// compute the total wall time
	totalWallTime= endWallTime-startWallTime;
	
	// compute the maximum wall time
	if(parallel) MPI_Reduce(&totalWallTime,&maximumWallTime,1,MPI_DOUBLE,MPI_MAX,0,WORLD);
	if(MASTER) printf("Maximum wall clock time %f s\n",maximumWallTime);	
	if (parallel) MPI_Finalize();
	return 0;
}

void synchronizeScalars(double** theOldArray, double** theNewArray) {

}

int computeClosestDivisibleInteger(int divisibleBy,int inputNumber) {
	double theRemainder = inputNumber % divisibleBy;
	while (theRemainder !=0) {
		theRemainder = ++inputNumber % divisibleBy;				
	}
	return inputNumber;
}

void setupBoundaryConditions(double** theArray,int arraySizeX, int arraySizeY) {
	/* set boundary conditions for ix, jy = 0 and ix, jy = n-1 */
	/* Here, we are using Dirichlet conditions */
	int ix, jy;
	double leftBC = 10, 
		rightBC = 10, 
		topBC = 20, 
		bottomBC = 20;
	
	/* setup the bottom and top BCs, jy = 0 and jy = n-1 or arraySizeY - 1 */
	for (ix = 0; ix < arraySizeX; ix++) {
		theArray[ix][0] = bottomBC;	//bottom BC
		theArray[ix][arraySizeY-1] = topBC;	//top BC
	}	
	
	/* setup the left and right BCs, ix = 0 and ix = arraySizeX - 1 */
	for (jy = 0; jy < arraySizeY; jy++) {
		theArray[0][jy] = leftBC;	//left BC
		theArray[arraySizeX-1][jy] = rightBC;	//right BC
	}
	
	/* set the values at the corner nodes as averages of both sides*/
	// bottom left
	theArray[0][0] = (leftBC + bottomBC)*0.5;
	// top left
	theArray[0][arraySizeY-1] = 0.5*(topBC + leftBC);
	// top right
	theArray[arraySizeX-1][arraySizeY-1] = 0.5*(topBC + rightBC);
	// bottom right
	theArray[arraySizeX-1][0] = 0.5*(bottomBC + rightBC);
}

void initializeInteriorNodes(double** theArray, int arraySizeX, int arraySizeY) {
	int ix, jy;
	for (ix = 1; ix <= arraySizeX - 2; ix++) {
		for (jy = 1; jy <= arraySizeY - 2; jy++) {
			theArray[ix][jy] = ix*jy;
		}
	}
}

void initializeArray(double** theArray, int arraySizeX, int arraySizeY) {
	setupBoundaryConditions(theArray,arraySizeX,arraySizeY);
	initializeInteriorNodes(theArray,arraySizeX,arraySizeY);	
}

double** make2DDoubleArray(int arraySizeX, int arraySizeY) {
	double** theArray;
	theArray = (double**) malloc(arraySizeX*sizeof(double*));
	for (int ix = 0; ix < arraySizeX; ix++) {
		theArray[ix] =(double*) malloc(arraySizeY*sizeof(double));
	}
	return theArray;
}

double* makeDoubleArray(int arraySize) {
	double* theArray;
	theArray = (double*) malloc(arraySize*sizeof(double));
	return theArray;
}

int* makeIntArray(int arraySize) {
	int* theArray;
	theArray = (int*) malloc(arraySize*sizeof(int));
	return theArray;
}

int** make2DIntArray(int arraySizeX, int arraySizeY) {
	int** theArray;
	theArray = (int**) malloc(arraySizeX*sizeof(int*));
	for (int ix = 0; ix < arraySizeX; ix++) {
		theArray[ix] =(int*) malloc(arraySizeY*sizeof(int));
	}
	return theArray;
}

void MPISetup(int argc, char* argv[]) {
	MPI_Init(&argc,&argv);
	MPI_Comm_size(WORLD,&numberOfProcessors);	
	MPI_Comm_rank(WORLD,&processorID);
	MPI_BuildDerivedTypes();
}

bool processUserInput(ProblemInfo &theProblemInfo) {
	int suggestedNumber;
	double remainder;

	printf("Enter the number of nodes in the x (or y) direction:");
	fflush(stdout);
	scanf("%d",&theProblemInfo.globalNumberOfNodes);	
	
	if (theProblemInfo.globalNumberOfNodes == 0) return 0;
	
	remainder = theProblemInfo.globalNumberOfNodes % numberOfProcessors;
	while (remainder !=0) {
		suggestedNumber = computeClosestDivisibleInteger(numberOfProcessors,theProblemInfo.globalNumberOfNodes);
		printf("Please enter a number of nodes that is divisible by the number of processors (try %i for example):",suggestedNumber);
		fflush(stdout);
		scanf("%d",&theProblemInfo.globalNumberOfNodes);			
		if (theProblemInfo.globalNumberOfNodes == 0) return 0;
		remainder = theProblemInfo.globalNumberOfNodes % numberOfProcessors;			
	}

	printf("Enter thermalDiffusivity:");
	fflush(stdout);
	scanf("%lf",&theProblemInfo.thermalDiffusivity);		
	
	printf("Enter end simulation time:");
	fflush(stdout);
	scanf("%lf",&theProblemInfo.endSimulationTime);		
	
	printf("Enter the Courant-Friedrichs-Lewy (CFL) number \n (For stability, the CFL number should be less than 0.5):");
	fflush(stdout);
	scanf("%lf",&theProblemInfo.CFL);

	/* compute the local number of nodes - each processor will need to add 2 shadow nodes in the x direction */
	theProblemInfo.numberOfNodesY = theProblemInfo.globalNumberOfNodes;
	theProblemInfo.numberOfNodesX = theProblemInfo.globalNumberOfNodes/numberOfProcessors + 2;

	return 1;
}



static void MPI_BuildProblemInfo(ProblemInfo p, MPI_Datatype* MPI_ProblemInfo) {
  int i;  
  int memoryBlockLengths[6];	
  MPI_Aint memoryOffset[6];	
  MPI_Datatype typeList[6];

  for (i = 0; i < 6; i++) {
	  memoryBlockLengths[i] = 1;	  
  }

  typeList[0] = MPI_INT;
  typeList[1] = MPI_INT;
  typeList[2] = MPI_INT;
  typeList[3] = MPI_DOUBLE;
  typeList[4] = MPI_DOUBLE;
  typeList[5] = MPI_DOUBLE;
	
  MPI_Address( &p.globalNumberOfNodes, &memoryOffset[0]);
  MPI_Address( &p.numberOfNodesX, &memoryOffset[1]);
  MPI_Address( &p.numberOfNodesY, &memoryOffset[2]);  
  MPI_Address( &p.endSimulationTime, &memoryOffset[3]);
  MPI_Address( &p.thermalDiffusivity, &memoryOffset[4]);
  MPI_Address( &p.CFL, &memoryOffset[5]);

  for ( i = 5; i >= 0; i--)
    memoryOffset[i] -= memoryOffset[0] ;
	
  MPI_Type_struct( 6, memoryBlockLengths, memoryOffset, typeList, MPI_ProblemInfo);
			
  MPI_Type_commit( MPI_ProblemInfo);
}

void MPI_BuildDerivedTypes() {
	ProblemInfo p;
	MPI_BuildProblemInfo(p, &MPI_ProblemInfo);
}